/*************************************************
*                                                *
*  EasyBMP Cross-Platform Windows Bitmap Library * 
*                                                *
*  Author: Paul Macklin                          *
*   email: pmacklin@math.uci.edu                 *
*                                                *
*    file: EasyBMP_BMP.h                         *
*    date: 5-2-2005                              *
* version: 0.63                                  *
*                                                *
* description: Defines BMP class                 *
*                                                *
*************************************************/

bool SafeFread( char* buffer, int size, int number, FILE* fp )
{
 if( !feof(fp) )
 {
  fread( buffer , size , number , fp );
  return true;
 }
 return false;
}

class BMP
{private:
 int BitDepth;
 int Width;
 int Height;

 public: 

 RGBApixel** Pixels;
 RGBApixel* Colors;

 int TellBitDepth( void );
 int TellWidth( void );
 int TellHeight( void );
 int TellNumberOfColors( void );
  
 BMP()
 {
  Width = 1;
  Height = 1;
  BitDepth = 24;
  Pixels = new RGBApixel* [Width];
  Pixels[0] = new RGBApixel [Height];
  Colors = NULL;
 }

 ~BMP()
 {
  int i=0;
  for(i=0;i<Width;i++)
  {
   delete [] Pixels[i];
  }
  delete [] Pixels;
  if( Colors )
  { delete Colors; }
 } 

 RGBApixel* operator()(int i, int j)
 {
  bool Warn = false;
  if( i >= Width ){ i = Width-1; Warn = true; }
  if( i < 0 ){ i = 0; Warn = true; }
  if( j >= Height ){ j = Height-1; Warn = true; }
  if( j < 0 ){ j = 0; Warn = true; }
  if( Warn )
  {
   cout << "EasyBMP Warning: Attempted to access non-existent pixel;\n"
        << "                 Truncating request to fit in domain [0,"
		<< Width-1 << "] x [0," << Height-1 << "].\n";
  }	
  
  return &(Pixels[i][j]);
 }
 
 void SetSize( int NewWidth, int NewHeight );
 void SetBitDepth( int NewDepth );
 void WriteToFile( char* FileName );
 void ReadFromFile( char* FileName );
};

int BMP::TellBitDepth( void )
{ return BitDepth; }

int BMP::TellHeight( void )
{ return Height; }

int BMP::TellWidth( void )
{ return Width; }

int BMP::TellNumberOfColors( void )
{
 int output = IntPow( 2, BitDepth );
 if( BitDepth == 32 )
 { output = IntPow( 2, 24 ); }
 return output;
}

void BMP::SetBitDepth( int NewDepth )
{
 if( NewDepth != 1 && NewDepth != 4 && 
     NewDepth != 8 && NewDepth != 24 && NewDepth != 32 )
 {
  cout << "EasyBMP Warning: User attempted to set unsupported bit depth " << NewDepth << ".\n"
       << "                 Bit depth remained unchanged at " << BitDepth << ".\n";
  return;
 }

 BitDepth = NewDepth;
 if( Colors )
 { delete Colors; }
 int NumberOfColors = IntPow( 2, BitDepth );
 if( BitDepth < 24 )
 { Colors = new RGBApixel [NumberOfColors]; }
 else
 { Colors = NULL; } 
 if( BitDepth < 24 )
 {
  CreateStandardColorTable( &Colors , BitDepth );
 }
 
 return;
}

void BMP::SetSize(int NewWidth , int NewHeight )
{
 if( NewWidth <= 0 || NewHeight <= 0 )
 {
  cout << "EasyBMP Warning: User attempted to set a non-positive width or height.\n"
       << "                 Size remained unchanged at " << Width << " x " << Height << ".\n";
  return;
 }

 int i=0, j=0; 

 for(i=0;i<Width;i++)
 {
  delete [] Pixels[i];
 }
 delete [] Pixels;

 Width = NewWidth;
 Height = NewHeight;
 Pixels = new RGBApixel* [ Width ]; 
 
 for(i=0; i<Width; i++)
 {
  Pixels[i] = new RGBApixel [ Height ];
 }
 
 for( i=0 ; i < Width ; i++)
 {
  for( j=0 ; j < Height ; j++ )
  {
   Pixels[i][j].Red = 255; 
   Pixels[i][j].Green = 255; 
   Pixels[i][j].Blue = 255; 
   Pixels[i][j].Alpha = 0;    
  }
 }

 return; 
}

void BMP::WriteToFile( char* FileName )
{
 FILE* fp = fopen( FileName, "wb" );
 if( fp == NULL )
 {
  cout << "EasyBMP Error: Cannot open file " << FileName << " for output.\n";
  fclose( fp );
  return;
 }
 
 WORD TempWORD;
 DWORD TempDWORD;
 
 // some preliminaries
 
 double dBytesPerPixel = ( (double) BitDepth ) / 8.0;
 double dBytesPerRow = dBytesPerPixel * (Width+0.0);
 dBytesPerRow = ceil(dBytesPerRow);
  
 int BytePaddingPerRow = 4 - ( (int) (dBytesPerRow) )% 4;
 if( BytePaddingPerRow == 4 )
 { BytePaddingPerRow = 0; } 
 
 double dActualBytesPerRow = dBytesPerRow + BytePaddingPerRow;
 
 double dTotalPixelBytes = Height * dActualBytesPerRow;
 
 double dPaletteSize = 0;
 if( BitDepth == 1 || BitDepth == 4 || BitDepth == 8 )
 { dPaletteSize = IntPow(2,BitDepth)*4.0; }
 
 double dTotalFileSize = 14 + 40 + dPaletteSize + dTotalPixelBytes;
 
 // write the file header 
 
 BMFH bmfh;
 bmfh.bfSize = (DWORD) dTotalFileSize; // bfSize
 bmfh.bfReserved1 = 0; // bfReserved1
 bmfh.bfReserved2 = 0; // bfReserved2
 bmfh.bfOffBits = (DWORD) (14+40+dPaletteSize); // bfOffBits 
 
 if( IsBigEndian() )
 {
  cout << "\tSwitching endianess for big-endian architecture ... \n";
  bmfh.SwitchEndianess();
 }
 
 fwrite( (char*) &(bmfh.bfType) , sizeof(WORD) , 1 , fp );
 fwrite( (char*) &(bmfh.bfSize) , sizeof(DWORD) , 1 , fp );
 fwrite( (char*) &(bmfh.bfReserved1) , sizeof(WORD) , 1 , fp );
 fwrite( (char*) &(bmfh.bfReserved2) , sizeof(WORD) , 1 , fp );
 fwrite( (char*) &(bmfh.bfOffBits) , sizeof(DWORD) , 1 , fp );
 
 // write the info header 
 
 BMIH bmih;
 bmih.biSize = 40;
 bmih.biWidth = Width;
 bmih.biHeight = Height;
 bmih.biPlanes = 1;
 bmih.biBitCount = BitDepth;
 bmih.biCompression = 0;
 bmih.biSizeImage = (DWORD) dTotalPixelBytes;
 bmih.biXPelsPerMeter = 0;
 bmih.biYPelsPerMeter = 0;
 bmih.biClrUsed = 0;
 bmih.biClrImportant = 0;
 
 if( IsBigEndian() )
 {
  cout << "\tSwitching endianess for big-endian architecture ... \n";
  bmih.SwitchEndianess();
 }
 
 fwrite( (char*) &(bmih.biSize) , sizeof(DWORD) , 1 , fp );
 fwrite( (char*) &(bmih.biWidth) , sizeof(DWORD) , 1 , fp );
 fwrite( (char*) &(bmih.biHeight) , sizeof(DWORD) , 1 , fp );
 fwrite( (char*) &(bmih.biPlanes) , sizeof(WORD) , 1 , fp );
 fwrite( (char*) &(bmih.biBitCount) , sizeof(WORD) , 1 , fp );
 fwrite( (char*) &(bmih.biCompression) , sizeof(DWORD) , 1 , fp );
 fwrite( (char*) &(bmih.biSizeImage) , sizeof(DWORD) , 1 , fp );
 fwrite( (char*) &(bmih.biXPelsPerMeter) , sizeof(DWORD) , 1 , fp );
 fwrite( (char*) &(bmih.biYPelsPerMeter) , sizeof(DWORD) , 1 , fp ); 
 fwrite( (char*) &(bmih.biClrUsed) , sizeof(DWORD) , 1 , fp);
 fwrite( (char*) &(bmih.biClrImportant) , sizeof(DWORD) , 1 , fp);
 
 // write the palette 
 if( BitDepth == 1 || BitDepth == 4 || BitDepth == 8 )
 {
  int NumberOfColors = IntPow(2,BitDepth);
  
  // if there is no palette, create one 
  if( !Colors )
  {
   if( !Colors )
   { Colors = new RGBApixel [NumberOfColors]; }
   CreateStandardColorTable( &Colors , BitDepth );
  }
   
  int n;
  for( n=0 ; n < NumberOfColors ; n++ )
  {
   fwrite( (char*) &(Colors[n]) , 4 , 1 , fp );
  }
 }
 
 // write the pixels 
 int i,j;
 if( BitDepth == 24 || BitDepth == 32 )
 {
  int MagicNumber = BitDepth / 8;
  for( j = Height-1 ; j >= 0 ; j-- )
  {
   for( i = 0 ; i < Width ; i++ )
   {
    fwrite( (char*) &(Pixels[i][j])  , MagicNumber , 1 , fp);     
   }
   // add any necessary padding to align to _ byte boundaries
   char* szTemp;
   szTemp = new char [BytePaddingPerRow];
   fwrite( szTemp , BytePaddingPerRow , 1 , fp );   
  }
 }
 
 if( BitDepth == 8 )
 {
  int DataBytes = Width;
  int PaddingBytes = ( 4 - DataBytes % 4 ) % 4;
 
  for( j = Height-1 ; j >= 0 ; j-- )
  {
   // write all row pixel data
   i=0;
   int WriteNumber = 0;
   while( WriteNumber < DataBytes )
   {
    int Index1 = 0;
	
    int k=0;
    int BestDistance = 999999;	
    if( i < Width )
    {
     // choose a best index.
     while( k < 256 && BestDistance > 0 )
     {
      int TempDistance = IntSquare( (int) Pixels[i][j].Red - (int) Colors[k].Red )
                       + IntSquare( (int) Pixels[i][j].Green - (int) Colors[k].Green ) 
                       + IntSquare( (int) Pixels[i][j].Blue - (int) Colors[k].Blue );
      if( TempDistance < BestDistance )
      {
       BestDistance = TempDistance;
       Index1 = k;
      }	  
      k++;
     }
     i++;	
    }
    BYTE RawBYTE = (BYTE) Index1;
    fwrite( (char*) &RawBYTE , 1, 1, fp);
    WriteNumber++;
   }
   // read any necessary row padding
   WriteNumber = 0;
   while( WriteNumber < PaddingBytes )
   {
    BYTE TempBYTE;
    fwrite( (char*) &TempBYTE , 1, 1, fp);
    WriteNumber++;
   }
  }
 }

 if( BitDepth == 4 )
 {
  int DataBytes = (int) ceil( Width / 2.0 );
  int PaddingBytes = ( 4 - DataBytes % 4 ) % 4;
 
  for( j = Height-1 ; j >= 0 ; j-- )
  {
   // write all row pixel data
   i=0;
   int WriteNumber = 0;
   while( WriteNumber < DataBytes )
   {
    int Index1 = 0;
    int Index2 = 0;
    
    int k;
    int BestDistance;	
    if( i < Width )
    {
     k=0;
     BestDistance = 999999;	
     // choose a best index.
     while( k < 16 )
     {
      int TempDistance = IntSquare( (int) Pixels[i][j].Red - (int) Colors[k].Red )
                       + IntSquare( (int) Pixels[i][j].Green - (int) Colors[k].Green )
                       + IntSquare( (int) Pixels[i][j].Blue - (int) Colors[k].Blue );
      if( TempDistance < BestDistance )
      {
       BestDistance = TempDistance;
       Index1 = k;
      }	  
      k++;
     }
     i++;	
    }
    
    if( i < Width )
    {
     k=0;
     BestDistance = 999999;	
     // choose a best index.
     while( k < 16 )
     {
      int TempDistance = IntSquare( (int) Pixels[i][j].Red - (int) Colors[k].Red )
                       + IntSquare( (int) Pixels[i][j].Green - (int) Colors[k].Green )
                       + IntSquare( (int) Pixels[i][j].Blue - (int) Colors[k].Blue );
      if( TempDistance < BestDistance )
      {
       BestDistance = TempDistance;
       Index2 = k;
      }	  
      k++;
     }
     i++;	
    }
    
    BYTE RawBYTE = (BYTE) (Index2 + Index1*16);
    fwrite( (char*) &RawBYTE , 1, 1, fp);
    WriteNumber++;
   }
   // read any necessary row padding
   WriteNumber = 0;
   while( WriteNumber < PaddingBytes )
   {
    BYTE TempBYTE;
    fwrite( (char*) &TempBYTE , 1, 1, fp);
    WriteNumber++;
   }
  }
 }

 if( BitDepth == 1 )
 {
  int DataBytes = (int) ceil( Width / 8.0 );
  int PaddingBytes = ( 4 - DataBytes % 4 ) % 4;
 
  for( j = Height-1 ; j >= 0 ; j-- )
  {
   // write all row pixel data
   i=0;
   int WriteNumber = 0;
   while( WriteNumber < DataBytes )
   {
    int Index1 = 0;
    int Index2 = 0;
    int Index3 = 0;
    int Index4 = 0;
    int Index5 = 0;
    int Index6 = 0;
    int Index7 = 0;
    int Index8 = 0;
	
    int k;
    int BestDistance;	
    if( i < Width )
    {
     k=0;
     BestDistance = 999999;	
     // choose a best index.
     while( k < 2 )
     {
      int TempDistance = IntSquare( (int) Pixels[i][j].Red - (int) Colors[k].Red )
                       + IntSquare( (int) Pixels[i][j].Green - (int) Colors[k].Green )
                       + IntSquare( (int) Pixels[i][j].Blue - (int) Colors[k].Blue );
      if( TempDistance < BestDistance )
      {
       BestDistance = TempDistance;
       Index1 = k;
      }	  
      k++;
     }
     i++;	
    }
              
    if( i < Width )
    {
     k=0;
     BestDistance = 999999;	
     // choose a best index.
     while( k < 2 )
     {
      int TempDistance = IntSquare( (int) Pixels[i][j].Red - (int) Colors[k].Red )
                       + IntSquare( (int) Pixels[i][j].Green - (int) Colors[k].Green )
                       + IntSquare( (int) Pixels[i][j].Blue - (int) Colors[k].Blue );
      if( TempDistance < BestDistance )
      {
       BestDistance = TempDistance;
       Index2 = k;
      }	  
      k++;
     }
     i++;	
    }
       	
    if( i < Width )
    {
     k=0;
     BestDistance = 999999;	
     // choose a best index.
     while( k < 2 )
     {
      int TempDistance = IntSquare( (int) Pixels[i][j].Red - (int) Colors[k].Red )
                       + IntSquare( (int) Pixels[i][j].Green - (int) Colors[k].Green )
                       + IntSquare( (int) Pixels[i][j].Blue - (int) Colors[k].Blue );
      if( TempDistance < BestDistance )
      {
       BestDistance = TempDistance;
       Index3 = k;
      }	  
      k++;
     }
     i++;	
    }
          	
    if( i < Width )
    {
     k=0;
     BestDistance = 999999;	
     // choose a best index.
     while( k < 2 )
     {
      int TempDistance = IntSquare( (int) Pixels[i][j].Red - (int) Colors[k].Red )
                       + IntSquare( (int) Pixels[i][j].Green - (int) Colors[k].Green )
                       + IntSquare( (int) Pixels[i][j].Blue - (int) Colors[k].Blue );
      if( TempDistance < BestDistance )
      {
       BestDistance = TempDistance;
       Index4 = k;
      }	  
      k++;
     }
     i++;	
    }
           
    if( i < Width )
    {
     k=0;
     BestDistance = 999999;	
     // choose a best index.
     while( k < 2 )
     {
      int TempDistance = IntSquare( (int) Pixels[i][j].Red - (int) Colors[k].Red )
                       + IntSquare( (int) Pixels[i][j].Green - (int) Colors[k].Green )
                       + IntSquare( (int) Pixels[i][j].Blue - (int) Colors[k].Blue );
      if( TempDistance < BestDistance )
      {
       BestDistance = TempDistance;
       Index5 = k;
      }	  
      k++;
     }
     i++;	
    }
         	
    if( i < Width )
    {
     k=0;
     BestDistance = 999999;	
     // choose a best index.
     while( k < 2 )
     {
      int TempDistance = IntSquare( (int) Pixels[i][j].Red - (int) Colors[k].Red )
                       + IntSquare( (int) Pixels[i][j].Green - (int) Colors[k].Green )
                       + IntSquare( (int) Pixels[i][j].Blue - (int) Colors[k].Blue );
      if( TempDistance < BestDistance )
      {
       BestDistance = TempDistance;
       Index6 = k;
      }	  
      k++;
     }
     i++;	
    }
        	
    if( i < Width )
    {
     k=0;
     BestDistance = 999999;	
     // choose a best index.
     while( k < 2 )
     {
      int TempDistance = IntSquare( (int) Pixels[i][j].Red - (int) Colors[k].Red )
                       + IntSquare( (int) Pixels[i][j].Green - (int) Colors[k].Green )
                       + IntSquare( (int) Pixels[i][j].Blue - (int) Colors[k].Blue );
      if( TempDistance < BestDistance )
      {
       BestDistance = TempDistance;
       Index7 = k;
      }	  
      k++;
     }
     i++;	
    }
        
    if( i < Width )
    {
     k=0;
     BestDistance = 999999;	
     // choose a best index.
     while( k < 2 )
     {
      int TempDistance = IntSquare( (int) Pixels[i][j].Red - (int) Colors[k].Red )
                       + IntSquare( (int) Pixels[i][j].Green - (int) Colors[k].Green )
                       + IntSquare( (int) Pixels[i][j].Blue - (int) Colors[k].Blue );
      if( TempDistance < BestDistance )
      {
       BestDistance = TempDistance;
       Index8 = k;
      }	  
      k++;
     }
     i++;	
    }
        
    BYTE RawBYTE = (BYTE) ( Index8    + Index7*2  + Index6*4  + Index5*8
                           +Index4*16 + Index3*32 + Index2*64 + Index1*128 );
    fwrite( (char*) &RawBYTE , 1, 1, fp);
    WriteNumber++;
   }
   // read any necessary row padding
   WriteNumber = 0;
   while( WriteNumber < PaddingBytes )
   {
    BYTE TempBYTE;
    fwrite( (char*) &TempBYTE , 1, 1, fp);
    WriteNumber++;
   }
  }
 }

 fclose(fp);
 return;
}

void BMP::ReadFromFile( char* FileName )
{ 
 FILE* fp = fopen( FileName, "rb" );
 if( fp == NULL )
 {
  cout << "EasyBMP Error: Cannot open file " << FileName << " for output.\n";
  SetBitDepth(1);
  SetSize(1,1);
  return;
 }
  
 WORD TempWORD;
 DWORD TempDWORD;
 
 // read the file header 
 
 BMFH bmfh;
 bool NotCorrupted = true;
 
 NotCorrupted = SafeFread( (char*) &(bmfh.bfType) , sizeof(WORD), 1, fp); // bfType
 
 bool IsBitmap = false;
 
 if( IsBigEndian() && bmfh.bfType == 16973 )
 { IsBitmap = true; }
 if( !IsBigEndian() && bmfh.bfType == 19778 )
 { IsBitmap = true; }
 
 if( !IsBitmap ) 
 {
  cout << "EasyBMP Error: " << FileName << " is not a Windows BMP file!\n"; 
  fclose( fp ); 
  return;
 }

 NotCorrupted = SafeFread( (char*) &(bmfh.bfSize) , sizeof(DWORD) , 1, fp); // bfSizse 
 NotCorrupted = SafeFread( (char*) &(bmfh.bfReserved1) , sizeof(WORD) , 1, fp); // bfReserved1
 NotCorrupted = SafeFread( (char*) &(bmfh.bfReserved2) , sizeof(WORD) , 1, fp); // bfReserved2
 NotCorrupted = SafeFread( (char*) &(bmfh.bfOffBits) , sizeof(DWORD) , 1 , fp); // bfOffbits
 
 if( IsBigEndian() ) 
 {
  cout << "\tSwitching endianess for big-endian architecture ... \n"; 
  bmfh.SwitchEndianess();
 }
  
 // read the info header

 BMIH bmih; 
 
 NotCorrupted = SafeFread( (char*) &(bmih.biSize) , sizeof(DWORD) , 1 , fp); // bmih.biSize
 NotCorrupted = SafeFread( (char*) &(bmih.biWidth) , sizeof(DWORD) , 1 , fp); // bmih.biWidth 
 NotCorrupted = SafeFread( (char*) &(bmih.biHeight) , sizeof(DWORD) , 1 , fp); // bmih.biHeight
 NotCorrupted = SafeFread( (char*) &(bmih.biPlanes) , sizeof(WORD) , 1, fp);  // bmih.biPlanes 
 NotCorrupted = SafeFread( (char*) &(bmih.biBitCount) , sizeof(WORD) , 1, fp); //   bmih.biBitCount 

 NotCorrupted = SafeFread( (char*) &(bmih.biCompression) , sizeof(DWORD) , 1 , fp); // bmih.biCompression
 NotCorrupted = SafeFread( (char*) &(bmih.biSizeImage) , sizeof(DWORD) , 1 , fp); // bmih.biSizeImage
 NotCorrupted = SafeFread( (char*) &(bmih.biXPelsPerMeter) , sizeof(DWORD) , 1 , fp); // bmih.biXPelsPerMeter
 NotCorrupted = SafeFread( (char*) &(bmih.biYPelsPerMeter) , sizeof(DWORD) , 1 , fp); // bmih.biYPelsPerMeter
 NotCorrupted = SafeFread( (char*) &(bmih.biClrUsed) , sizeof(DWORD) , 1 , fp); // bmih.biClrUsed
 NotCorrupted = SafeFread( (char*) &(bmih.biClrImportant) , sizeof(DWORD) , 1 , fp); // bmih.biClrImportant
 
 if( IsBigEndian() ) 
 {
  cout << "\tSwitching endianess for big-endian architecture ... \n"; 
  bmih.SwitchEndianess();
 }

 // a safety catch: if any of the header information didn't read properly, abort
 // future idea: check to see if at least most of self-consistent
  
 if( !NotCorrupted )
 {
  cout << "EasyBMP Error: " << FileName << " is obviously corrupted. \n";
  SetSize(1,1);
  SetBitDepth(1);
  fclose(fp);
  return;
 } 
 
 if(    BitDepth != 1  && BitDepth != 4 && BitDepth != 8 
     && BitDepth != 24 && BitDepth != 32 )
 {
  cout << "EasyBMP Error: " << FileName << " has unrecognized bit depth. \n";
  SetSize(1,1);
  SetBitDepth(1);
  fclose(fp);
  return;
 }
 
 // set the size

 if( (int) bmih.biWidth <= 0 || (int) bmih.biHeight <= 0 ) 
 {
  cout << "EasyBMP Error: " << FileName << " has a non-positive width or height. \n";
  SetSize(1,1);
  SetBitDepth(1);
  fclose(fp);
  return;
 }

 
 SetSize( (int) bmih.biWidth , (int) bmih.biHeight );
 SetBitDepth( (int) bmih.biBitCount );
  
 // some preliminaries
 
 double dBytesPerPixel = ( (double) BitDepth ) / 8.0;
 double dBytesPerRow = dBytesPerPixel * (Width+0.0);
 dBytesPerRow = ceil(dBytesPerRow);
  
 int BytePaddingPerRow = 4 - ( (int) (dBytesPerRow) )% 4;
 if( BytePaddingPerRow == 4 )
 { BytePaddingPerRow = 0; }  
 
 // if < 24 bits, read the palette

 if( BitDepth == 1 || BitDepth == 4 || BitDepth == 8 )
 {
  int NumberOfColors = IntPow(2,BitDepth);
  Colors = new RGBApixel [NumberOfColors];
  int n;
  for( n=0; n < NumberOfColors ; n++ )
  {
   SafeFread( (char*) &(Colors[n]) , 4 , 1 , fp);     
  }
 }
  
 //read the pixels

 int i,j;
 if( BitDepth == 24 || BitDepth == 32 )
 {
  int SizePerRead = BitDepth / 8;
  for( j = Height-1 ; j >= 0 ; j-- )
  {
   for( i = 0 ; i < Width ; i++ )
   {
    SafeFread( (char*) &(Pixels[i][j])  , SizePerRead , 1 , fp);     
   }
   // read any necessary padding to align to 4 byte boundaries
   char* szTemp;
   szTemp = new char [BytePaddingPerRow];
   SafeFread( szTemp , BytePaddingPerRow , 1 , fp );   
  }
 }
 
 if( BitDepth == 8 )
 {
  int DataBytes = Width;
  int PaddingBytes = ( 4 - DataBytes % 4 ) % 4;
 
  for( j = Height-1 ; j >= 0 ; j-- )
  {
   // read all row pixel data
   i=0;
   int ReadNumber = 0;
   while( ReadNumber < DataBytes )
   {
    BYTE TempBYTE;
    SafeFread( (char*) &TempBYTE , 1, 1, fp);
    ReadNumber++;
    int RawIndex = (int)TempBYTE;
    int Index1 = RawIndex;
    (Pixels[i][j]).Red = Colors[Index1].Red;
    (Pixels[i][j]).Green = Colors[Index1].Green;
    (Pixels[i][j]).Blue = Colors[Index1].Blue;
    (Pixels[i][j]).Alpha = Colors[Index1].Alpha;
    i++;
   }
   // read any necessary row padding
   ReadNumber = 0;
   while( ReadNumber < PaddingBytes )
   {
    BYTE TempBYTE;
    SafeFread( (char*) &TempBYTE , 1, 1, fp);
    ReadNumber++;
   }
  }
 }
 
 if( BitDepth == 4 )
 {
  int DataBytes = (int) ceil( Width*0.5 );
  int PaddingBytes = ( 4 - DataBytes % 4 ) % 4;
 
  for( j = Height-1 ; j >= 0 ; j-- )
  {
   // read all row pixel data
   i=0;
   int ReadNumber = 0;
   while( ReadNumber < DataBytes )
   {
    BYTE TempBYTE;
    SafeFread( (char*) &TempBYTE , 1, 1, fp);
    ReadNumber++;
    int RawIndex = (int)TempBYTE;
    int Index2 = RawIndex % 16;
    int Index1 = (RawIndex - Index2)/16;
    (Pixels[i][j]).Red = Colors[Index1].Red;
    (Pixels[i][j]).Green = Colors[Index1].Green;
    (Pixels[i][j]).Blue = Colors[Index1].Blue;
    (Pixels[i][j]).Alpha = Colors[Index1].Alpha;
    i++;
    if( i < Width )
    {
     (Pixels[i][j]).Red = Colors[Index2].Red;
     (Pixels[i][j]).Green = Colors[Index2].Green;
     (Pixels[i][j]).Blue = Colors[Index2].Blue;
     (Pixels[i][j]).Alpha = Colors[Index2].Alpha;
     i++;
    }	
   }
   
   // read any necessary row padding
   ReadNumber = 0;
   while( ReadNumber < PaddingBytes )
   {
    BYTE TempBYTE;
    SafeFread( (char*) &TempBYTE , 1, 1, fp);
    ReadNumber++;
   }
  }  
 }
 
 if( BitDepth == 1 )
 {
  int DataBytes = (int) ceil( Width/8.0 );
  int PaddingBytes = ( 4 - DataBytes % 4 ) % 4;
 
  for( j = Height-1 ; j >= 0 ; j-- )
  {
   // read all row pixel data
   i=0;
   int ReadNumber = 0;
   while( ReadNumber < DataBytes )
   {
    BYTE TempBYTE;
    SafeFread( (char*) &TempBYTE , 1, 1, fp);
    ReadNumber++;
    int RawIndex = (int)TempBYTE;
    int Index8 = RawIndex % 2;
    RawIndex = (RawIndex - Index8)/2;
    int Index7 = RawIndex % 2;
    RawIndex = (RawIndex - Index7)/2;
    int Index6 = RawIndex % 2;
    RawIndex = (RawIndex - Index6)/2;
    int Index5 = RawIndex % 2;
    RawIndex = (RawIndex - Index5)/2;
    int Index4 = RawIndex % 2;
    RawIndex = (RawIndex - Index4)/2;
    int Index3 = RawIndex % 2;
    RawIndex = (RawIndex - Index3)/2;
    int Index2 = RawIndex % 2;
    RawIndex = (RawIndex - Index2)/2;
    int Index1 = RawIndex;
    
    int Index = Index1;
    (Pixels[i][j]).Red = Colors[Index].Red;
    (Pixels[i][j]).Green = Colors[Index].Green;
    (Pixels[i][j]).Blue = Colors[Index].Blue;
    (Pixels[i][j]).Alpha = Colors[Index].Alpha;
    i++;
    if( i < Width )
    {Index = Index2;
     (Pixels[i][j]).Red = Colors[Index].Red;
     (Pixels[i][j]).Green = Colors[Index].Green;
     (Pixels[i][j]).Blue = Colors[Index].Blue;
     (Pixels[i][j]).Alpha = Colors[Index].Alpha;
     i++;
    }	
    if( i < Width )
    {Index = Index3;
     (Pixels[i][j]).Red = Colors[Index].Red;
     (Pixels[i][j]).Green = Colors[Index].Green;
     (Pixels[i][j]).Blue = Colors[Index].Blue;
     (Pixels[i][j]).Alpha = Colors[Index].Alpha;
     i++;
    }	
    if( i < Width )
    {Index = Index4;
     (Pixels[i][j]).Red = Colors[Index].Red;
     (Pixels[i][j]).Green = Colors[Index].Green;
     (Pixels[i][j]).Blue = Colors[Index].Blue;
     (Pixels[i][j]).Alpha = Colors[Index].Alpha;
     i++;
    }	
    if( i < Width )
    {Index = Index5;
     (Pixels[i][j]).Red = Colors[Index].Red;
     (Pixels[i][j]).Green = Colors[Index].Green;
     (Pixels[i][j]).Blue = Colors[Index].Blue;
     (Pixels[i][j]).Alpha = Colors[Index].Alpha;
     i++;
    }	
    if( i < Width )
    {Index = Index6;
     (Pixels[i][j]).Red = Colors[Index].Red;
     (Pixels[i][j]).Green = Colors[Index].Green;
     (Pixels[i][j]).Blue = Colors[Index].Blue;
     (Pixels[i][j]).Alpha = Colors[Index].Alpha;
     i++;
    }	
    if( i < Width )
    {Index = Index7;
     (Pixels[i][j]).Red = Colors[Index].Red;
     (Pixels[i][j]).Green = Colors[Index].Green;
     (Pixels[i][j]).Blue = Colors[Index].Blue;
     (Pixels[i][j]).Alpha = Colors[Index].Alpha;
     i++;
    }	
    if( i < Width )
    {Index = Index8;
     (Pixels[i][j]).Red = Colors[Index].Red;
     (Pixels[i][j]).Green = Colors[Index].Green;
     (Pixels[i][j]).Blue = Colors[Index].Blue;
     (Pixels[i][j]).Alpha = Colors[Index].Alpha;
     i++;
    }		
   }
   // read any necessary row padding
   ReadNumber = 0;
   while( ReadNumber < PaddingBytes )
   {
    BYTE TempBYTE;
    SafeFread( (char*) &TempBYTE , 1, 1, fp);
    ReadNumber++;
   }
  }   
 }
 fclose(fp);
 return;
}
